/* SPDX-License-Identifier: GPL-2.0 WITH Linux-syscall-note OR MIT */

/*
 * (C) COPYRIGHT 2020-2023 ARM Limited or its affiliates. All rights reserved.
 */

/*
 * Public interface for Mali Assign Module
 */

#ifndef _MALI_GPU_ASSIGN_H_
#define _MALI_GPU_ASSIGN_H_

#define MALI_PTM_RESOURCE_GROUP_COUNT (4)
#define MALI_PTM_ACCESS_WINDOW_COUNT (16)
#define MALI_PTM_PARTITION_COUNT (4)
#define MALI_PTM_SLICES_COUNT (8)

#endif /* _MALI_GPU_ASSIGN_H_ */
